import os

REDIS_HOST = os.environ.get('REDIS_HOST', 'careful-redis')
# TODO: should be renamed to ELASTIC_URL
ELASTIC_HOST = f"http://{os.environ.get('ELASTIC_HOST','careful-elastic')}:9200"

REDIS_OPENAI_RATE_LIMIT_LOCK_KEY = os.environ.get('REDIS_OPENAI_RATE_LIMIT_LOCK__KEY', 'openai_rate_limit_lock')

REDIS_OPENAI_RATE_LIMIT_LAST_FILL_TIME_KEY = os.environ.get('REDIS_OPENAI_RATE_LIMIT_LAST_FILL_TIME_KEY', 'openai_rate_limit_last_fill_time')

REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY = os.environ.get('REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY', 'openai_rate_limit_tokens')


import unittest
import threading
import asyncio
import time
from unittest.mock import patch

import numpy as np

from patent_chart.redis_utils import r
from patent_chart import settings
from patent_chart import generator
from patent_chart import ranking
from patent_chart.requests_utils import DistributedFixedWindowRateLimiter

# NOTE: REDIS_HOST should be set to the host of the redis server accessible to the test runner
class TestDistributedRateLimiter(unittest.IsolatedAsyncioTestCase):
    class _MockServer:
        def __init__(self):
            self._requests = []
            self._tokens = []
            self._timestamps = []

        async def request(self, request, tokens):
            self._requests.append(request)
            self._tokens.append(tokens)
            self._timestamps.append(time.monotonic())

        def get_requests(self):
            return self._requests
        
        def get_tokens(self):
            return self._tokens
        
        def get_timestamps(self):
            return self._timestamps
    
    def setUp(self):
        r.delete(settings.REDIS_OPENAI_RATE_LIMIT_LAST_FILL_TIME_KEY)
        r.delete(settings.REDIS_OPENAI_RATE_LIMIT_TOKENS_KEY)
        r.delete(settings.REDIS_OPENAI_RATE_LIMIT_LOCK_KEY)

    def start_make_request_loop(self, server):
        asyncio.run(self._make_requests(server))

    async def _make_requests(self, server):
        async with DistributedFixedWindowRateLimiter(10, 1) as limiter:
            for i in range(3):
                tokens = 8
                await limiter.wait_for_tokens(tokens)
                await server.request(i, tokens)
    
    def test_distributed_rate_limiter(self):
        # spawn three threads that each make a series of requests to a mock server
        server = self._MockServer()
        threades = []
        for _ in range(3):
            thread = threading.Thread(target=self.start_make_request_loop, args=(server,))
            thread.start()
            threades.append(thread)

        # Wait for all threades to finish
        for thread in threades:
            thread.join()
        # Ensure that all requests have been made at the end of the test
        expected = []
        for i in range(3):
            expected.extend([i] * 3)

        self.assertEqual(
            sorted(server.get_requests()),
            expected
        )

        timestamps = server.get_timestamps()
        tokens = server.get_tokens()
        bins = np.arange(timestamps[0], timestamps[-1], 1)
        tokens_per_bin = np.zeros(len(bins) - 1)
        i = 1
        for tok, time in zip(tokens, timestamps):
            if time > bins[i]:
                i += 1
            tokens_per_bin[i - 1] += tok

        print(tokens_per_bin)
        self.assertTrue(np.all(tokens_per_bin <= 10))


class TestRanking(unittest.TestCase):
    def test_ranking(self):
        claim_element = 'this is a brown dog'
        generated_passages = [
            generator.GeneratedPassage(
                claim_element_id=0,
                prior_art_source_id=0,
                text='this is a brown dog',
                model_id='model',
            ),
            generator.GeneratedPassage(
                claim_element_id=0,
                prior_art_source_id=0,
                text='lebron james is a basketball player',
                model_id='model',
            ),
            generator.GeneratedPassage(
                claim_element_id=0,
                prior_art_source_id=0,
                text='this is a brown cat',
                model_id='model',
            ),
        ]

        ranked_passages = ranking.rank_passages(generated_passages, claim_element, ranking_model_version='text-embedding-3-large')
        self.assertEqual(
            [p.ranking for p in ranked_passages],
            [1, 3, 2]
        )


import random
import asyncio
import time
from logging import getLogger
from typing import Optional

import numpy as np
# from sentence_transformers import SentenceTransformer
# See https://gitlab.com/bloomsday-labs/ulysses/-/issues/47
import openai
import tiktoken

from .data_structures import GeneratedPassage
from .requests_utils import DistributedFixedWindowRateLimiter
from .utils import num_tokens_from_messages

logger = getLogger(__name__)

# _cached_sentence_transformer_model = None

# sentence_transformer_models = [
#     'all-MiniLM-L6-v2'
# ]
openai_models = [
    'text-embedding-3-large'
]

def num_tokens_from_string(string: str, encoding_name: str) -> int:
    """Returns the number of tokens in a text string."""
    encoding = tiktoken.get_encoding(encoding_name)
    num_tokens = len(encoding.encode(string))
    return num_tokens

# def get_cached_sentence_transformer_model(model_name):
#     global _model
#     if _cached_sentence_transformer_model is None:
#         _cached_sentence_transformer_model = SentenceTransformer(model_name)
#     return _cached_sentence_transformer_model

def openai_embedding_request_with_retry(model_name, input, backoff_factor=2, backoff_override=None, n_retries=3, **kwargs) -> list[list[float]]:
    """Send a chat completion request to OpenAI with retries."""
    for i in range(1, 1 + n_retries):
        try:
            embeddings = openai.Embedding.create(
                input=input,
                model=model_name,
                **kwargs
            )
            return [e['embedding'] for e in embeddings['data']]
        # except openai.OpenAIError as e:
        except openai.error.OpenAIError as e:
            logger.error(f"OpenAIError: {e}")
            if backoff_override is not None:
                jitter = random.uniform(-0.10, 0.10)
                time.sleep(backoff_override * (1 + jitter))
            else:
                jitter = random.uniform(-0.10, 0.10) 
                time.sleep(backoff_factor**i * (1 + jitter))
            continue
        except Exception as e:
            raise e
    raise Exception(f"Failed to send embedding request to OpenAI after {n_retries} retries.")

async def aopenai_embedding_request_with_retry(model_name, input, backoff_factor=2, backoff_override=None, n_retries=3, **kwargs) -> list[list[float]]:
    """Send a chat completion request to OpenAI with retries."""
    for i in range(1, 1 + n_retries):
        try:
            embeddings = await openai.Embedding.acreate(
                input=input,
                model=model_name,
                **kwargs
            )
            return [e['embedding'] for e in embeddings['data']]
        # except openai.OpenAIError as e:
        except openai.error.OpenAIError as e:
            logger.error(f"OpenAIError: {e}")
            if backoff_override is not None:
                jitter = random.uniform(-0.10, 0.10)
                await asyncio.sleep(backoff_override * (1 + jitter))
            else:
                jitter = random.uniform(-0.10, 0.10) 
                await asyncio.sleep(backoff_factor**i * (1 + jitter))
            continue
        except Exception as e:
            raise e
    raise Exception(f"Failed to send embedding request to OpenAI after {n_retries} retries.")

class RankingModel:
    def __init__(self, model):
        self.model_name = model
        # if self.model_name not in sentence_transformer_models and self.model_name not in openai_models:
        #     raise ValueError(f'Invalid model name: {model}')
        if self.model_name not in openai_models:
            raise ValueError(f'Invalid model name: {model}')

    # @property
    # def openai_client(self):
    #     if not hasattr(self, '_open_ai_client'):
    #         self._openai_client = openai.OpenAI()
    #     return self._openai_client

    def encode(self, chunks: list[str] | str):
        # if self.model_name in sentence_transformer_models:
        #     return get_cached_sentence_transformer_model(self.model_name).encode(chunks)
        # elif self.model_name in openai_models:
        if self.model_name in openai_models:
            if isinstance(chunks, str):
                chunks = [chunks]
            # embeddings = self.openai_client.embeddings.create(
            #     input=chunks,
            #     model=self.model_name,
            #     dimensions=256,
            # )
            return openai_embedding_request_with_retry(self.model_name, chunks, dimensions=256)
        
    async def aencode(self, chunks: list[str] | str):
        if self.model_name in openai_models:
            if isinstance(chunks, str):
                chunks = [chunks]
            return await aopenai_embedding_request_with_retry(self.model_name, chunks, dimensions=256)


def cosine_similarity(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))

def find_most_similar_chunks(embeddings: np.ndarray, ref: np.ndarray):
    similarities: list[np.ndarray] = []
    for embedding in embeddings:
        similarities.append(cosine_similarity(ref, embedding))

    similarities = np.vstack(similarities).reshape(-1)
    return np.argsort(similarities)[::-1]

def rank_passages(passages: list[GeneratedPassage], claim_element: str, ranking_model_version: str):
    """
    Mutates the ranking attribute of each passage in the list
    """
    model = RankingModel(ranking_model_version)
    claim_element_embedding = model.encode(claim_element)
    passage_embeddings = model.encode([passage.text for passage in passages])
    sorted_indices = find_most_similar_chunks(passage_embeddings, claim_element_embedding)
    
    for i, sorted_idx in enumerate(sorted_indices):
        passages[sorted_idx].ranking = i + 1
        passages[sorted_idx].ranking_model_version = ranking_model_version

    return passages

async def arank_passages(passages: list[GeneratedPassage], claim_element: str, ranking_model_version: str):
    """
    Mutates the ranking attribute of each passage in the list
    """
    model = RankingModel(ranking_model_version)
    claim_element_embedding = model.encode(claim_element)
    passage_embeddings = await model.aencode([passage.text for passage in passages])
    sorted_indices = find_most_similar_chunks(passage_embeddings, claim_element_embedding)
    
    for i, sorted_idx in enumerate(sorted_indices):
        passages[sorted_idx].ranking = i + 1
        passages[sorted_idx].ranking_model_version = ranking_model_version

    return passages


import requests
import pathlib
import time
import typing
from logging import getLogger

from bs4 import BeautifulSoup

from patent_chart.data_structures import GoogleParsedPatent, PatentUniqueID

logger = getLogger(__name__)


class GoogleParseError(Exception):
    pass


def google_request_with_retry(url, max_retries=3):
    for i in range(1, max_retries + 1):
        response = requests.get(url)
        if response.status_code == 200:
            return response
        else:
            logger.warning('Failed to request %s. Response code: %s. Error: %s', url, response.status_code, response.text)
            logger.warning('Retrying %s', url)
            time.sleep(i**2)
    return None

def parse_google_patent(patent_unique_id: PatentUniqueID) ->  GoogleParsedPatent | None:
    google_id = f'{patent_unique_id.country_code}{patent_unique_id.patent_number}{patent_unique_id.kind_code or ""}'
    url = f"https://patents.google.com/patent/{google_id}/en"

    logger.info('Requesting %s', url)

    # Send an HTTP GET request to fetch the webpage
    response = google_request_with_retry(url)

    # Parse the HTML content of the page using BeautifulSoup
    soup = BeautifulSoup(response.text, 'html.parser')

    specification = soup.find('section', {'itemprop': 'description'})
    claims = soup.find('section', {'itemprop': 'claims'})
    title = soup.find('td', {'itemprop': 'title'})

    if specification is None or claims is None:
        logger.warning('Failed to parse %s', url)
        return None
    
    title_text = None
    if title is not None:
        title_text = title.text.strip(' \n')

    return GoogleParsedPatent(
        unique_id=patent_unique_id,
        title=title_text,
        specification=specification.text.strip(' \n'),
        claims=claims.text.strip(' \n')
    )

# if __name__ == '__main__':
#     package_dir = pathlib.Path(__file__).parents[1]
#     us_patents_dir = package_dir / 'us_patents_1980-2020'
#     for pdf_path in us_patents_dir.glob('*.pdf'):
#         unique_id = parser.parse_patent_unique_id_from_pdf_path(pdf_path)
#         if unique_id is not None:
#             google_unique_id = f'{unique_id.country_code}{unique_id.patent_number}{unique_id.kind_code or ""}'
#             result = parse_google_patent(google_unique_id)
#             if result is not None:
#                 specification, claims = result
#                 assert specification != ''
#                 assert claims != ''
#             else:
#                 print(f'Failed to parse {unique_id}')
#         else:
#             print(f'Failed to parse {pdf_path}')


from elasticsearch import Elasticsearch

from . import settings
from . import parser

client = Elasticsearch(settings.ELASTIC_HOST)

XML_USPTO_POST_2001 = 'xml_uspto_post_2001'

claim_embedding_scheme = {
    "properties": {
        "patent_doc_id": {
            "type": "keyword"
        },
        "claim_id": {
            "type": "keyword"
        },
        "embedding": {
            "type": "dense_vector",
            "dims": 1024
        }
    }
}

specification_embedding_scheme = {
    "properties": {
        "patent_doc_id": {
            "type": "keyword"
        },
        "paragraph_id": {
            "type": "keyword"
        },
        "embedding": {
            "type": "dense_vector",
            "dims": 1024
        }
    }
}

patent_scheme = {
    "properties": {
        "unique_id": {
            "properties": {
                "patent_number": {
                    "type": "keyword"
                },
                "country_code": {
                    "type": "keyword"
                },
                "kind_code": {
                    "type": "keyword"
                },
            }
        },
        "claims": {
            "type": "text"
        },
        'claims_format': {
            'properties': {
                'data_format': {
                    'type': 'keyword'
                },
                'tree_structure': {
                    'type': 'keyword'
                }
            }
        },
        'specification': {
            "type": "text"
        },
        'specification_format': {
            'properties': {
                'data_format': {
                    'type': 'keyword'
                },
                'tree_structure': {
                    'type': 'keyword'
                }
            }
        },
        'title': {
            'type': 'text'
        },
        'summary': {
            'type': 'text'
        },
        'summary_model_version': {
            'type': 'keyword'
        },
        'short_summary': {
            'type': 'text'
        },
        'short_summary_model_version': {
            'type': 'keyword'
        },
        'topics': {
            'type': 'keyword'
        },
        'topics_model_version': {
            'type': 'keyword'
        },
        'text': {
            'type': 'text'
        },
        'text_format': {
            'type': 'keyword'
        },
    }
}

def index_patent(unique_id, text, title, text_format=None):
    body = {
        "unique_id": {
            "patent_number": unique_id.patent_number,
            "kind_code": unique_id.kind_code,
            "country_code": unique_id.country_code
        },
        "text": text,
        "title": title
    }

    if text_format:
        body['text_format'] = text_format
    else:
        body['text_format'] = XML_USPTO_POST_2001
        
    client.index(index='patents', id=str(unique_id), body=body)
    return str(unique_id)


def index_claim_embedding(patent_doc_id, embedding, claim_id=None):
    body = {
        "patent_doc_id": patent_doc_id,
        "embedding": embedding
    }
    if claim_id:
        body['claim_id'] = claim_id

    client.index(index='patents_claim_embedding', body=body)

def index_specification_embedding(patent_doc_id, embedding, paragraph_id=None):
    body = {
        "patent_doc_id": patent_doc_id,
        "embedding": embedding
    }
    if paragraph_id:
        body['paragraph_id'] = paragraph_id
    client.index(index='patents_specification_embedding', body=body)

def get_patent_summary(patent: parser.GoogleParsedPatent) -> str:
    response = client.get(index='patents', id=str(patent.unique_id))
    return response['_source'].get('summary')

def update_patent_by_id(patent: parser.GoogleParsedPatent, **kwargs):
    client.update(index='patents', id=str(patent.unique_id), body={'doc': kwargs})